from django.dispatch import Signal

# providing_args=['request', 'refresh_token']
refresh_token_revoked = Signal()

# providing_args=['request', 'refresh_token', 'refresh_token_issued']
refresh_token_rotated = Signal()



from django.contrib.auth import get_user_model

import graphene

from . import mixins
from .decorators import create_token
from .refresh_token.relay import DeleteRefreshTokenCookie, Revoke

__all__ = [
    "JWTMutation",
    "Create",
    "Verify",
    "Refresh",
    "Revoke",
    "DeleteRefreshTokenCookie",
]


class JWTMutation(mixins.CreateJWTMixin, graphene.ClientIDMutation):
    class Meta:
        abstract = True

    @classmethod
    def Field(cls, *args, **kwargs):
        cls._meta.arguments["input"]._meta.fields.update(
            {
                get_user_model().USERNAME_FIELD: graphene.InputField(
                    graphene.String,
                    required=True,
                ),
                "password": graphene.InputField(graphene.String, required=True),
            },
        )
        return super().Field(*args, **kwargs)

    @classmethod
    @create_token
    def mutate_and_get_payload(cls, root, info, **kwargs):
        return cls.create(root, info, **kwargs)


class Create(mixins.CreateMixin, JWTMutation):
    """Create JSON Web Token mutation"""


class Verify(mixins.VerifyMixin, graphene.ClientIDMutation):
    class Input:
        token = graphene.String()

    @classmethod
    def mutate_and_get_payload(cls, *args, **kwargs):
        return cls.verify(*args, **kwargs)


class Refresh(mixins.RefreshMixin, graphene.ClientIDMutation):
    class Input(mixins.RefreshMixin.Fields):
        """Refresh Input"""

    @classmethod
    def mutate_and_get_payload(cls, *args, **kwargs):
        return cls.refresh(*args, **kwargs)


class DeleteJWTCookie(
    mixins.DeleteJWTCookieMixin,
    graphene.ClientIDMutation,
):
    @classmethod
    def mutate_and_get_payload(cls, *args, **kwargs):
        return cls.delete_cookie(*args, **kwargs)



from django.contrib.admin import SimpleListFilter
from django.utils.translation import gettext_lazy as _

__all__ = [
    "ExpiredFilter",
    "RevokedFilter",
]


class BooleanListFilter(SimpleListFilter):
    def lookups(self, request, model_admin):
        return (
            ("yes", _("Yes")),
            ("no", _("No")),
        )


class ExpiredFilter(BooleanListFilter):
    title = _("Expired")
    parameter_name = "expired"

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.expired().filter(expired=True)

        if self.value() == "no":
            return queryset.expired().filter(expired=False)


class RevokedFilter(BooleanListFilter):
    title = _("Revoked")
    parameter_name = "revoked"

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(revoked__isnull=False)

        if self.value() == "no":
            return queryset.filter(revoked__isnull=True)



from datetime import timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    "JWT_ALGORITHM": "HS256",
    "JWT_AUDIENCE": None,
    "JWT_ISSUER": None,
    "JWT_LEEWAY": 0,
    "JWT_SECRET_KEY": settings.SECRET_KEY,
    "JWT_PUBLIC_KEY": None,
    "JWT_PRIVATE_KEY": None,
    "JWT_VERIFY": True,
    "JWT_VERIFY_EXPIRATION": False,
    "JWT_EXPIRATION_DELTA": timedelta(seconds=60 * 5),
    "JWT_ALLOW_REFRESH": True,
    "JWT_REFRESH_EXPIRATION_DELTA": timedelta(days=7),
    "JWT_REFRESH_TOKEN_LONG_RUNNING": False,
    "JWT_REFRESH_TOKEN_MODEL": "refresh_token.RefreshToken",
    "JWT_REFRESH_TOKEN_N_BYTES": 20,
    "JWT_REFRESH_TOKEN_REUSE": False,
    "JWT_AUTH_HEADER_NAME": "HTTP_AUTHORIZATION",
    "JWT_AUTH_HEADER_PREFIX": "JWT",
    "JWT_ALLOW_ARGUMENT": False,
    "JWT_ARGUMENT_NAME": "token",
    "JWT_ENCODE_HANDLER": "django_jwt.utils.jwt_encode",
    "JWT_DECODE_HANDLER": "django_jwt.utils.jwt_decode",
    "JWT_PAYLOAD_HANDLER": "django_jwt.utils.jwt_payload",
    "JWT_PAYLOAD_GET_USERNAME_HANDLER": (
        lambda payload: payload.get(get_user_model().USERNAME_FIELD)
    ),
    "JWT_GET_USER_BY_NATURAL_KEY_HANDLER": "django_jwt.utils.get_user_by_natural_key",
    "JWT_REFRESH_EXPIRED_HANDLER": "django_jwt.utils.refresh_has_expired",
    "JWT_GET_REFRESH_TOKEN_HANDLER": (
        "django_jwt.refresh_token.utils.get_refresh_token_by_model"
    ),
    "JWT_ALLOW_ANY_HANDLER": "django_jwt.middleware.allow_any",
    "JWT_ALLOW_ANY_CLASSES": (),
    "JWT_CSRF_ROTATION": False,
    "JWT_HIDE_TOKEN_FIELDS": False,
    "JWT_COOKIE_NAME": "JWT",
    "JWT_REFRESH_TOKEN_COOKIE_NAME": "JWT-refresh-token",
    "JWT_COOKIE_SECURE": False,
    "JWT_COOKIE_PATH": "/",
    "JWT_COOKIE_DOMAIN": None,
    "JWT_COOKIE_SAMESITE": None,
}

IMPORT_STRINGS = (
    "JWT_ENCODE_HANDLER",
    "JWT_DECODE_HANDLER",
    "JWT_PAYLOAD_HANDLER",
    "JWT_PAYLOAD_GET_USERNAME_HANDLER",
    "JWT_GET_USER_BY_NATURAL_KEY_HANDLER",
    "JWT_REFRESH_EXPIRED_HANDLER",
    "JWT_GET_REFRESH_TOKEN_HANDLER",
    "JWT_ALLOW_ANY_HANDLER",
    "JWT_ALLOW_ANY_CLASSES",
)


def perform_import(value, setting_name):
    if isinstance(value, str):
        return import_from_string(value, setting_name)
    if isinstance(value, (list, tuple)):
        return [import_from_string(item, setting_name) for item in value]
    return value


def import_from_string(value, setting_name):
    try:
        return import_string(value)
    except ImportError as e:
        msg = (
            f"Could not import `{value}` for JWT setting `{setting_name}`."
            f"{e.__class__.__name__}: {e}."
        )
        raise ImportError(msg)


class JWTSettings:
    def __init__(self, defaults, import_strings):
        self.defaults = defaults
        self.import_strings = import_strings
        self._cached_attrs = set()

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid setting: `{attr}`")

        value = self.user_settings.get(attr, self.defaults[attr])

        if attr == "JWT_ALLOW_ANY_CLASSES":
            value = list(value) + [
                "django_jwt.mixins.JWTMixin",
                "django_jwt.mixins.VerifyMixin",
                "django_jwt.refresh_token.mixins.RevokeMixin",
            ]

        if attr in self.import_strings:
            value = perform_import(value, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, value)
        return value

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "DJANGO_JWT", {})
        return self._user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)

        self._cached_attrs.clear()

        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


def reload_settings(*args, **kwargs):
    setting = kwargs["setting"]

    if setting == "DJANGO_JWT":
        jwt_settings.reload()


setting_changed.connect(reload_settings)

jwt_settings = JWTSettings(DEFAULTS, IMPORT_STRINGS)



# Generated by Django 2.1.2 on 2018-10-16 20:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="RefreshToken",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "token",
                    models.CharField(
                        editable=False, max_length=255, verbose_name="token"
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "revoked",
                    models.DateTimeField(blank=True, null=True, verbose_name="revoked"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="refresh_tokens",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "Refresh token",
                "verbose_name_plural": "Refresh tokens",
                "abstract": False,
            },
        ),
        migrations.AlterUniqueTogether(
            name="refreshtoken",
            unique_together={("token", "revoked")},
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("refresh_token", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="refreshtoken",
            options={
                "verbose_name": "refresh token",
                "verbose_name_plural": "refresh tokens",
            },
        ),
    ]



from django.contrib.auth import get_user_model

from .shortcuts import get_user_by_token
from .utils import get_credentials

UserModel = get_user_model()


class JWTBackend:
    def authenticate(self, request=None, **kwargs):
        if request is None or getattr(request, "_jwt_token_auth", False):
            return None

        token = get_credentials(request, **kwargs)

        if token is not None:
            return get_user_by_token(token, request)

        return None

    def get_user(self, user_id):
        try:
            return UserModel._default_manager.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


class JWTRESTBackend:
    def authenticate(self, request=None, **kwargs):
        if request is None or getattr(request, "_jwt_token_auth", False):
            return None

        token = get_credentials(request, **kwargs)

        if token is not None:
            return get_user_by_token(token, request)

        return None

    def get_user(self, user_id):
        try:
            return UserModel._default_manager.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


class JWTGraphQLBackend:
    def authenticate(self, request=None, **kwargs):
        if request is None or getattr(request, "_jwt_token_auth", False):
            return None

        token = get_credentials(request, **kwargs)

        if token is not None:
            return get_user_by_token(token, request)

        return None

    def get_user(self, user_id):
        try:
            return UserModel._default_manager.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


from calendar import timegm
from datetime import datetime

from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _

import jwt

from . import exceptions
from .settings import jwt_settings


def jwt_payload(user, context=None):
    username = user.get_username()

    if hasattr(username, "pk"):
        username = username.pk

    exp = datetime.utcnow() + jwt_settings.JWT_EXPIRATION_DELTA

    payload = {
        user.USERNAME_FIELD: username,
        "exp": timegm(exp.utctimetuple()),
    }

    if jwt_settings.JWT_ALLOW_REFRESH:
        payload["origIat"] = timegm(datetime.utcnow().utctimetuple())

    if jwt_settings.JWT_AUDIENCE is not None:
        payload["aud"] = jwt_settings.JWT_AUDIENCE

    if jwt_settings.JWT_ISSUER is not None:
        payload["iss"] = jwt_settings.JWT_ISSUER

    return payload


def jwt_encode(payload, context=None):
    return jwt.encode(
        payload,
        jwt_settings.JWT_PRIVATE_KEY or jwt_settings.JWT_SECRET_KEY,
        jwt_settings.JWT_ALGORITHM,
    )


def jwt_decode(token, context=None):
    return jwt.decode(
        token,
        jwt_settings.JWT_PUBLIC_KEY or jwt_settings.JWT_SECRET_KEY,
        options={
            "verify_exp": jwt_settings.JWT_VERIFY_EXPIRATION,
            "verify_aud": jwt_settings.JWT_AUDIENCE is not None,
            "verify_signature": jwt_settings.JWT_VERIFY,
        },
        leeway=jwt_settings.JWT_LEEWAY,
        audience=jwt_settings.JWT_AUDIENCE,
        issuer=jwt_settings.JWT_ISSUER,
        algorithms=[jwt_settings.JWT_ALGORITHM],
    )


def get_http_authorization(request):
    auth = request.META.get(jwt_settings.JWT_AUTH_HEADER_NAME, "").split()
    prefix = jwt_settings.JWT_AUTH_HEADER_PREFIX

    if len(auth) != 2 or auth[0].lower() != prefix.lower():
        return request.COOKIES.get(jwt_settings.JWT_COOKIE_NAME)
    return auth[1]


def get_token_argument(request, **kwargs):
    if jwt_settings.JWT_ALLOW_ARGUMENT:
        input_fields = kwargs.get("input")

        if isinstance(input_fields, dict):
            kwargs = input_fields

        return kwargs.get(jwt_settings.JWT_ARGUMENT_NAME)
    return None


def get_credentials(request, **kwargs):
    return get_token_argument(request, **kwargs) or get_http_authorization(request)


def get_payload(token, context=None):
    try:
        payload = jwt_settings.JWT_DECODE_HANDLER(token, context)
    except jwt.ExpiredSignatureError:
        raise exceptions.JWTExpired()
    except jwt.DecodeError:
        raise exceptions.JWTError(_("Error decoding signature"))
    except jwt.InvalidTokenError:
        raise exceptions.JWTError(_("Invalid token"))
    return payload


def get_user_by_natural_key(username):
    UserModel = get_user_model()
    try:
        return UserModel._default_manager.get_by_natural_key(username)
    except UserModel.DoesNotExist:
        return None


def get_user_by_payload(payload):
    username = jwt_settings.JWT_PAYLOAD_GET_USERNAME_HANDLER(payload)

    if not username:
        raise exceptions.JWTError(_("Invalid payload"))

    user = jwt_settings.JWT_GET_USER_BY_NATURAL_KEY_HANDLER(username)

    if user is not None and not getattr(user, "is_active", True):
        raise exceptions.JWTError(_("User is disabled"))
    return user


def refresh_has_expired(orig_iat, context=None):
    exp = orig_iat + jwt_settings.JWT_REFRESH_EXPIRATION_DELTA.total_seconds()
    return timegm(datetime.utcnow().utctimetuple()) > exp


def set_cookie(response, key, value, expires):
    kwargs = {
        "expires": expires,
        "httponly": True,
        "secure": jwt_settings.JWT_COOKIE_SECURE,
        "path": jwt_settings.JWT_COOKIE_PATH,
        "domain": jwt_settings.JWT_COOKIE_DOMAIN,
        "samesite": jwt_settings.JWT_COOKIE_SAMESITE,
    }
    
    response.set_cookie(key, value, **kwargs)


def delete_cookie(response, key):
    response.delete_cookie(
        key,
        path=jwt_settings.JWT_COOKIE_PATH,
        domain=jwt_settings.JWT_COOKIE_DOMAIN,
    )



from django.contrib.auth import get_user_model

import graphene

from . import mixins
from .decorators import create_token
from .refresh_token.mutations import DeleteRefreshTokenCookie, Revoke

__all__ = [
    "JWTMutation",
    "Create",
    "Verify",
    "Refresh",
    "Revoke",
    "DeleteRefreshTokenCookie",
]


class JWTMutation(mixins.CreateJWTMixin, graphene.Mutation):
    class Meta:
        abstract = True

    @classmethod
    def Field(cls, *args, **kwargs):
        cls._meta.arguments.update(
            {
                get_user_model().USERNAME_FIELD: graphene.String(required=True),
                "password": graphene.String(required=True),
            },
        )
        return super().Field(*args, **kwargs)

    @classmethod
    @create_token
    def mutate(cls, root, info, **kwargs):
        return cls.create(root, info, **kwargs)


class Create(mixins.CreateMixin, JWTMutation):
    """Obtain JSON Web Token mutation"""


class Verify(mixins.VerifyMixin, graphene.Mutation):
    class Arguments:
        token = graphene.String()

    @classmethod
    def mutate(cls, *args, **kwargs):
        return cls.verify(*args, **kwargs)


class Refresh(mixins.RefreshMixin, graphene.Mutation):
    class Arguments(mixins.RefreshMixin.Fields):
        """Refresh Arguments"""

    @classmethod
    def mutate(cls, *arg, **kwargs):
        return cls.refresh(*arg, **kwargs)


class DeleteJWTCookie(mixins.DeleteJWTCookieMixin, graphene.Mutation):
    @classmethod
    def mutate(cls, *args, **kwargs):
        return cls.delete_cookie(*args, **kwargs)



try:
    from graphql.execution.execute import GraphQLResolveInfo
except ImportError:
    from graphql.execution.base import ResolveInfo as GraphQLResolveInfo  # noqa


def get_operation_name(operation):
    if hasattr(operation, "value"):
        return operation.value
    return operation



import binascii
import os
from calendar import timegm

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ..settings import jwt_settings
from . import managers, signals


class AbstractRefreshToken(models.Model):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="refresh_tokens",
        verbose_name=_("user"),
    )
    token = models.CharField(_("token"), max_length=255, editable=False)
    created = models.DateTimeField(_("created"), auto_now_add=True)
    revoked = models.DateTimeField(_("revoked"), null=True, blank=True)

    objects = managers.RefreshTokenQuerySet.as_manager()

    class Meta:
        abstract = True
        verbose_name = _("refresh token")
        verbose_name_plural = _("refresh tokens")
        unique_together = ("token", "revoked")

    def __str__(self):
        return self.token

    def save(self, *args, **kwargs):
        if not self.token:
            self.token = self._cached_token = self.generate_token()

        super().save(*args, **kwargs)

    def generate_token(self):
        return binascii.hexlify(
            os.urandom(jwt_settings.JWT_REFRESH_TOKEN_N_BYTES),
        ).decode()

    def get_token(self):
        if hasattr(self, "_cached_token"):
            return self._cached_token
        return self.token

    def is_expired(self, request=None):
        orig_iat = timegm(self.created.timetuple())
        return jwt_settings.JWT_REFRESH_EXPIRED_HANDLER(orig_iat, request)

    def revoke(self, request=None):
        self.revoked = timezone.now()
        self.save(update_fields=["revoked"])

        signals.refresh_token_revoked.send(
            sender=AbstractRefreshToken,
            request=request,
            refresh_token=self,
        )

    def reuse(self, request=None):
        self.token = ""
        self.created = timezone.now()
        self.save(update_fields=["token", "created"])


class RefreshToken(AbstractRefreshToken):
    """RefreshToken default model"""



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class RefreshTokenConfig(AppConfig):
    name = "django_jwt.refresh_token"
    verbose_name = _("Refresh token")



from . import relay
from .mutations import (
    DeleteJWTCookie,
    DeleteRefreshTokenCookie,
    JWTMutation,
    Create,
    Refresh,
    Revoke,
    Verify,
)

__all__ = [
    "relay",
    "JWTMutation",
    "Create",
    "Verify",
    "Refresh",
    "Revoke",
    "DeleteJWTCookie",
    "DeleteRefreshTokenCookie",
]

__version__ = "0.3.4"
